@@ -1,22 +1,60 @@
+// Copyright (C) 2004-2023 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #ifndef PDF_INTERPRET_H
 #define PDF_INTERPRET_H
 
 #include "mupdf/pdf/font.h"
 #include "mupdf/pdf/resource.h"
+#include "mupdf/pdf/document.h"
 
-typedef struct pdf_csi_s pdf_csi;
-typedef struct pdf_gstate_s pdf_gstate;
-typedef struct pdf_processor_s pdf_processor;
+typedef struct pdf_gstate pdf_gstate;
+typedef struct pdf_processor pdf_processor;
 
 void *pdf_new_processor(fz_context *ctx, int size);
+pdf_processor *pdf_keep_processor(fz_context *ctx, pdf_processor *proc);
 void pdf_close_processor(fz_context *ctx, pdf_processor *proc);
 void pdf_drop_processor(fz_context *ctx, pdf_processor *proc);
 
-struct pdf_processor_s
+struct pdf_processor
 {
+	int refs;
+
+	/* close the processor. Also closes any chained processors. */
 	void (*close_processor)(fz_context *ctx, pdf_processor *proc);
 	void (*drop_processor)(fz_context *ctx, pdf_processor *proc);
 
+	/* At any stage, we can have one set of resources in place.
+	 * This function gives us a set of resources to use. We remember
+	 * any previous set on a stack, so we can pop back to it later.
+	 * Our responsibility (as well as remembering it for our own use)
+	 * is to pass either it, or a filtered version of it onto any
+	 * chained processor. */
+	void (*push_resources)(fz_context *ctx, pdf_processor *proc, pdf_obj *res);
+	/* Pop the resources stack. This must be passed on to any chained
+	 * processors. This returns a pointer to the resource dict just
+	 * popped by the deepest filter. The caller inherits this reference. */
+	pdf_obj *(*pop_resources)(fz_context *ctx, pdf_processor *proc);
+
 	/* general graphics state */
 	void (*op_w)(fz_context *ctx, pdf_processor *proc, float linewidth);
 	void (*op_j)(fz_context *ctx, pdf_processor *proc, int linejoin);
@@ -30,7 +68,7 @@ struct pdf_processor_s
 	void (*op_gs_BM)(fz_context *ctx, pdf_processor *proc, const char *blendmode);
 	void (*op_gs_ca)(fz_context *ctx, pdf_processor *proc, float alpha);
 	void (*op_gs_CA)(fz_context *ctx, pdf_processor *proc, float alpha);
-	void (*op_gs_SMask)(fz_context *ctx, pdf_processor *proc, pdf_obj *smask, pdf_obj *page_resources, float *bc, int luminosity);
+	void (*op_gs_SMask)(fz_context *ctx, pdf_processor *proc, pdf_obj *smask, float *bc, int luminosity);
 	void (*op_gs_end)(fz_context *ctx, pdf_processor *proc);
 
 	/* special graphics state */
@@ -84,9 +122,9 @@ struct pdf_processor_s
 
 	/* text showing */
 	void (*op_TJ)(fz_context *ctx, pdf_processor *proc, pdf_obj *array);
-	void (*op_Tj)(fz_context *ctx, pdf_processor *proc, char *str, int len);
-	void (*op_squote)(fz_context *ctx, pdf_processor *proc, char *str, int len);
-	void (*op_dquote)(fz_context *ctx, pdf_processor *proc, float aw, float ac, char *str, int len);
+	void (*op_Tj)(fz_context *ctx, pdf_processor *proc, char *str, size_t len);
+	void (*op_squote)(fz_context *ctx, pdf_processor *proc, char *str, size_t len);
+	void (*op_dquote)(fz_context *ctx, pdf_processor *proc, float aw, float ac, char *str, size_t len);
 
 	/* type 3 fonts */
 	void (*op_d0)(fz_context *ctx, pdf_processor *proc, float wx, float wy);
@@ -110,10 +148,10 @@ struct pdf_processor_s
 	void (*op_k)(fz_context *ctx, pdf_processor *proc, float c, float m, float y, float k);
 
 	/* shadings, images, xobjects */
-	void (*op_BI)(fz_context *ctx, pdf_processor *proc, fz_image *image);
+	void (*op_BI)(fz_context *ctx, pdf_processor *proc, fz_image *image, const char *colorspace_name);
 	void (*op_sh)(fz_context *ctx, pdf_processor *proc, const char *name, fz_shade *shade);
 	void (*op_Do_image)(fz_context *ctx, pdf_processor *proc, const char *name, fz_image *image);
-	void (*op_Do_form)(fz_context *ctx, pdf_processor *proc, const char *name, pdf_obj *form, pdf_obj *page_resources);
+	void (*op_Do_form)(fz_context *ctx, pdf_processor *proc, const char *name, pdf_obj *form);
 
 	/* marked content */
 	void (*op_MP)(fz_context *ctx, pdf_processor *proc, const char *tag);
@@ -140,7 +178,7 @@ struct pdf_processor_s
 	int hidden;
 };
 
-struct pdf_csi_s
+typedef struct
 {
 	/* input */
 	pdf_document *doc;
@@ -158,33 +196,19 @@ struct pdf_csi_s
 	pdf_obj *obj;
 	char name[256];
 	char string[256];
-	int string_len;
+	size_t string_len;
 	int top;
 	float stack[32];
-};
+} pdf_csi;
 
 /* Functions to set up pdf_process structures */
 
-/*
-	pdf_new_run_processor: Create a new "run" processor. This maps
-	from PDF operators to fz_device level calls.
-
-	dev: The device to which the resulting device calls are to be
-	sent.
-
-	ctm: The initial transformation matrix to use.
-
-	usage: A NULL terminated string that describes the 'usage' of
-	this interpretation. Typically 'View', though 'Print' is also
-	defined within the PDF reference manual, and others are possible.
-
-	gstate: The initial graphics state.
-*/
-pdf_processor *pdf_new_run_processor(fz_context *ctx, fz_device *dev, fz_matrix ctm, const char *usage, pdf_gstate *gstate, fz_default_colorspaces *default_cs);
+pdf_processor *pdf_new_run_processor(fz_context *ctx, pdf_document *doc, fz_device *dev, fz_matrix ctm, int struct_parent, const char *usage, pdf_gstate *gstate, fz_default_colorspaces *default_cs, fz_cookie *cookie);
 
 /*
-	pdf_new_buffer_processor: Create a buffer processor. This
-	collects the incoming PDF operator stream into an fz_buffer.
+	Create a buffer processor.
+
+	This collects the incoming PDF operator stream into an fz_buffer.
 
 	buffer: The (possibly empty) buffer to which operators will be
 	appended.
@@ -195,7 +219,7 @@ pdf_processor *pdf_new_run_processor(fz_context *ctx, fz_device *dev, fz_matrix
 pdf_processor *pdf_new_buffer_processor(fz_context *ctx, fz_buffer *buffer, int ahxencode);
 
 /*
-	pdf_new_output_processor: Create an output processor. This
+	Create an output processor. This
 	sends the incoming PDF operator stream to an fz_output stream.
 
 	out: The output stream to which operators will be sent.
@@ -205,12 +229,108 @@ pdf_processor *pdf_new_buffer_processor(fz_context *ctx, fz_buffer *buffer, int
 */
 pdf_processor *pdf_new_output_processor(fz_context *ctx, fz_output *out, int ahxencode);
 
-/*
-	pdf_new_filter_processor: Create a filter processor. This
-	filters the PDF operators it is fed, and passes them down
-	(with some changes) to the child filter.
+typedef struct pdf_filter_options pdf_filter_options;
 
-	The changes made by the filter are:
+/*
+	Create a filter processor. This filters the PDF operators
+	it is fed, and passes them down (with some changes) to the
+	child filter.
+
+	chain: The child processor to which the filtered operators
+	will be fed.
+
+	The options field contains a pointer to a structure with
+	filter specific options in.
+*/
+typedef pdf_processor *(pdf_filter_factory_fn)(fz_context *ctx, pdf_document *doc, pdf_processor *chain, int struct_parents, fz_matrix transform, pdf_filter_options *options, void *factory_options);
+
+/*
+	A pdf_filter_factory is a pdf_filter_factory_fn, plus the options
+	needed to instantiate it.
+*/
+typedef struct
+{
+	pdf_filter_factory_fn *filter;
+	void *options;
+} pdf_filter_factory;
+
+/*
+	recurse: Filter resources recursively.
+
+	instance_forms: Always recurse on XObject Form resources, but will
+	create a new instance of each XObject Form that is used, filtered
+	individually.
+
+	ascii: If true, escape all binary data in the output.
+
+	no_update: If true, do not update the document at the end.
+
+	opaque: Opaque value that is passed to the complete function.
+
+	complete: A function called at the end of processing.
+	This allows the caller to insert some extra content after
+	all other content.
+
+	filters: Pointer to an array of filter factory/options.
+	The array is terminated by an entry with a NULL factory pointer.
+	Operators will be fed into the filter generated from the first
+	factory function in the list, and from there go to the filter
+	generated from the second factory in the list etc.
+*/
+struct pdf_filter_options
+{
+	int recurse;
+	int instance_forms;
+	int ascii;
+	int no_update;
+
+	void *opaque;
+	void (*complete)(fz_context *ctx, fz_buffer *buffer, void *arg);
+
+	pdf_filter_factory *filters;
+};
+
+typedef enum
+{
+	FZ_CULL_PATH_FILL,
+	FZ_CULL_PATH_STROKE,
+	FZ_CULL_PATH_FILL_STROKE,
+	FZ_CULL_CLIP_PATH,
+	FZ_CULL_GLYPH,
+	FZ_CULL_IMAGE,
+	FZ_CULL_SHADING
+} fz_cull_type;
+
+/*
+	image_filter: A function called to assess whether a given
+	image should be removed or not.
+
+	text_filter: A function called to assess whether a given
+	character should be removed or not.
+
+	after_text_object: A function called after each text object.
+	This allows the caller to insert some extra content if
+	desired.
+
+	culler: A function called to see whether each object should
+	be culled or not.
+*/
+typedef struct
+{
+	void *opaque;
+	fz_image *(*image_filter)(fz_context *ctx, void *opaque, fz_matrix ctm, const char *name, fz_image *image);
+	int (*text_filter)(fz_context *ctx, void *opaque, int *ucsbuf, int ucslen, fz_matrix trm, fz_matrix ctm, fz_rect bbox);
+	void (*after_text_object)(fz_context *ctx, void *opaque, pdf_document *doc, pdf_processor *chain, fz_matrix ctm);
+	int (*culler)(fz_context *ctx, void *opaque, fz_rect bbox, fz_cull_type type);
+}
+pdf_sanitize_filter_options;
+
+/*
+	A sanitize filter factory.
+
+	sopts = pointer to pdf_sanitize_filter_options.
+
+	The changes made by a filter generated from this are:
 
 	* No operations are allowed to change the top level gstate.
 	Additional q/Q operators are inserted to prevent this.
@@ -228,52 +348,70 @@ pdf_processor *pdf_new_output_processor(fz_context *ctx, fz_output *out, int ahx
 
 	The net graphical effect of the filtered operator stream
 	should be identical to the incoming operator stream.
-
-	chain: The child processor to which the filtered operators
-	will be fed.
-
-	old_res: The incoming resource dictionary.
-
-	new_res: An (initially empty) resource dictionary that will
-	be populated by copying entries from the old dictionary to
-	the new one as they are used. At the end therefore, this
-	contains exactly those resource objects actually required.
-
 */
-pdf_processor *pdf_new_filter_processor(fz_context *ctx, pdf_document *doc, pdf_processor *chain, pdf_obj *old_res, pdf_obj *new_res);
+pdf_processor *pdf_new_sanitize_filter(fz_context *ctx, pdf_document *doc, pdf_processor *chain, int struct_parents, fz_matrix transform, pdf_filter_options *options, void *sopts);
 
-typedef int (pdf_text_filter_fn)(fz_context *ctx, void *opaque, int *ucsbuf, int ucslen, fz_matrix trm, fz_rect bbox);
+pdf_obj *pdf_filter_xobject_instance(fz_context *ctx, pdf_obj *old_xobj, pdf_obj *page_res, fz_matrix ctm, pdf_filter_options *options, pdf_cycle_list *cycle_up);
 
-typedef void (pdf_after_text_object_fn)(fz_context *ctx, void *opaque, pdf_document *doc, pdf_processor *chain, fz_matrix ctm);
+void pdf_processor_push_resources(fz_context *ctx, pdf_processor *proc, pdf_obj *res);
+
+pdf_obj *pdf_processor_pop_resources(fz_context *ctx, pdf_processor *proc);
 
 /*
-	pdf_new_filter_processor_with_text_filter: Create a filter
-	processor with a filter function for text. This filters the
-	PDF operators it is fed, and passes them down (with some
-	changes) to the child filter.
+	opaque: Opaque value that is passed to all the filter functions.
 
-	See pdf_new_filter_processor for documentation.
+	color_rewrite: function pointer called to rewrite a color
+		On entry:
+			*cs = reference to a pdf object representing the colorspace.
 
-	text_filter: A function called to assess whether a given
-	character should be removed or not.
+			*n = number of color components
 
-	after_text_object: A function to be called after each text object.
-	This allows the caller to insert some extra content if
-	required.
+			color = *n color values.
 
-	text_filter_opaque: Opaque value to be passed to the
-	text_filter function.
+		On exit:
+			*cs either the same (for no change in colorspace) or
+			updated to be a new one. Reference must be dropped, and
+			a new kept reference returned!
+
+			*n = number of color components (maybe updated)
+
+			color = *n color values (maybe updated)
+
+	image_rewrite: function pointer called to rewrite an image
+		On entry:
+			*image = reference to an fz_image.
+
+		On exit:
+			*image either the same (for no change) or updated
+			to be a new one. Reference must be dropped, and a
+			new kept reference returned.
 */
-pdf_processor *
-pdf_new_filter_processor_with_text_filter(fz_context *ctx, pdf_document *doc, pdf_processor *chain, pdf_obj *old_rdb, pdf_obj *new_rdb, pdf_text_filter_fn *text_filter, pdf_after_text_object_fn *after, void *text_filter_opaque);
+typedef struct
+{
+	void *opaque;
+	void (*color_rewrite)(fz_context *ctx, void *opaque, pdf_obj **cs, int *n, float color[FZ_MAX_COLORS]);
+	void (*image_rewrite)(fz_context *ctx, void *opaque, fz_image **image);
+	pdf_shade_recolorer *shade_rewrite;
+} pdf_color_filter_options;
 
-/* Functions to actually process annotations, glyphs and general stream objects */
-void pdf_process_contents(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_obj *obj, pdf_obj *res, fz_cookie *cookie);
-void pdf_process_annot(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_page *page, pdf_annot *annot, fz_cookie *cookie);
+pdf_processor *
+pdf_new_color_filter(fz_context *ctx, pdf_document *doc, pdf_processor *chain, int struct_parents, fz_matrix transform, pdf_filter_options *options, void *copts);
+
+/*
+	Functions to actually process annotations, glyphs and general stream objects.
+*/
+void pdf_process_contents(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_obj *obj, pdf_obj *res, fz_cookie *cookie, pdf_obj **out_res);
+void pdf_process_annot(fz_context *ctx, pdf_processor *proc, pdf_annot *annot, fz_cookie *cookie);
 void pdf_process_glyph(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_obj *resources, fz_buffer *contents);
 
+/*
+	Function to process a contents stream without handling the resources.
+	The caller is responsible for pushing/popping the resources.
+*/
+void pdf_process_raw_contents(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_obj *rdb, pdf_obj *stmobj, fz_cookie *cookie);
+
 /* Text handling helper functions */
-typedef struct pdf_text_state_s
+typedef struct
 {
 	float char_space;
 	float word_space;
@@ -285,7 +423,7 @@ typedef struct pdf_text_state_s
 	float rise;
 } pdf_text_state;
 
-typedef struct pdf_text_object_state_s
+typedef struct
 {
 	fz_text *text;
 	fz_rect text_bbox;
