@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2022 Artifex Software, Inc.
+// Copyright (C) 2004-2023 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -151,7 +151,7 @@ build_compression_params(fz_context *ctx, pdf_obj *f, pdf_obj *p, fz_compression
 		pdf_obj *ct = pdf_dict_get(ctx, p, PDF_NAME(ColorTransform));
 
 		params->type = FZ_IMAGE_JPEG;
-		params->u.jpeg.color_transform = (ct ? pdf_to_int(ctx, ct) : -1);
+		params->u.jpeg.color_transform = (ct ? pdf_to_int(ctx, ct) : -2);
 	}
 	else if (pdf_name_eq(ctx, f, PDF_NAME(RunLengthDecode)) || pdf_name_eq(ctx, f, PDF_NAME(RL)))
 	{
@@ -515,22 +515,30 @@ pdf_load_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 	return buf;
 }
 
-static int
-pdf_guess_filter_length(int len, const char *filter)
+static size_t
+pdf_guess_filter_length(size_t len, const char *filter)
 {
-	if (len < 0)
-		len = 0;
+	size_t nlen = len;
+
+	/* First ones get smaller, no overflow check required. */
 	if (!strcmp(filter, "ASCIIHexDecode"))
 		return len / 2;
-	if (!strcmp(filter, "ASCII85Decode"))
+	else if (!strcmp(filter, "ASCII85Decode"))
 		return len * 4 / 5;
+
 	if (!strcmp(filter, "FlateDecode"))
-		return len * 3;
-	if (!strcmp(filter, "RunLengthDecode"))
-		return len * 3;
-	if (!strcmp(filter, "LZWDecode"))
-		return len * 2;
-	return len;
+		nlen = len * 3;
+	else if (!strcmp(filter, "RunLengthDecode"))
+		nlen = len * 3;
+	else if (!strcmp(filter, "LZWDecode"))
+		nlen = len * 2;
+
+	/* Live with a bad estimate - we'll malloc up as we go, but
+	 * it's probably destined to fail anyway. */
+	if (nlen < len)
+		return len;
+
+	return nlen;
 }
 
 /* Check if an entry has a cached stream and return whether it is directly
@@ -586,11 +594,12 @@ can_reuse_buffer(fz_context *ctx, pdf_xref_entry *entry, fz_compression_params *
 }
 
 static fz_buffer *
-pdf_load_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compression_params *params, int *truncated)
+pdf_load_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compression_params *params, int *truncated, size_t worst_case)
 {
 	fz_stream *stm = NULL;
 	pdf_obj *dict, *obj;
-	int i, len, n;
+	int i, n;
+	size_t len;
 	fz_buffer *buf;
 
 	fz_var(buf);
@@ -607,7 +616,14 @@ pdf_load_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compressio
 	dict = pdf_load_object(ctx, doc, num);
 	fz_try(ctx)
 	{
-		len = pdf_dict_get_int(ctx, dict, PDF_NAME(Length));
+		int64_t ilen = pdf_dict_get_int64(ctx, dict, PDF_NAME(Length));
+		if (ilen < 0)
+			ilen = 0;
+		len = (size_t)ilen;
+		/* In 32 bit builds, we might find a length being too
+		 * large for a size_t. */
+		if ((int64_t)len != ilen)
+			fz_throw(ctx, FZ_ERROR_GENERIC, "Stream too large");
 		obj = pdf_dict_get(ctx, dict, PDF_NAME(Filter));
 		len = pdf_guess_filter_length(len, pdf_to_name(ctx, obj));
 		n = pdf_array_len(ctx, obj);
@@ -627,10 +643,7 @@ pdf_load_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compressio
 
 	fz_try(ctx)
 	{
-		if (truncated)
-			buf = fz_read_best(ctx, stm, len, truncated);
-		else
-			buf = fz_read_all(ctx, stm, len);
+		buf = fz_read_best(ctx, stm, len, truncated, worst_case);
 	}
 	fz_always(ctx)
 	{
@@ -647,17 +660,17 @@ pdf_load_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compressio
 fz_buffer *
 pdf_load_stream_number(fz_context *ctx, pdf_document *doc, int num)
 {
-	return pdf_load_image_stream(ctx, doc, num, NULL, NULL);
+	return pdf_load_image_stream(ctx, doc, num, NULL, NULL, 0);
 }
 
 fz_compressed_buffer *
-pdf_load_compressed_stream(fz_context *ctx, pdf_document *doc, int num)
+pdf_load_compressed_stream(fz_context *ctx, pdf_document *doc, int num, size_t worst_case)
 {
 	fz_compressed_buffer *bc = fz_malloc_struct(ctx, fz_compressed_buffer);
 
 	fz_try(ctx)
 	{
-		bc->buffer = pdf_load_image_stream(ctx, doc, num, &bc->params, NULL);
+		bc->buffer = pdf_load_image_stream(ctx, doc, num, &bc->params, NULL, worst_case);
 	}
 	fz_catch(ctx)
 	{
